#ifndef DLIB_REVISION_H
// Version:  19.24
// Date:     Sun 08 May 2022 10:30:32 AM EDT
// Git Changeset ID:  6097093ab329fcd19aed03a8fe67949f6971a65d
#define DLIB_MAJOR_VERSION  19
#define DLIB_MINOR_VERSION  24
#define DLIB_PATCH_VERSION  0
#endif
